from ._better_print import better_print
from ._check_params_decorator import check_params


@check_params
def better_input(*args, sep: str = ' ', delay: float = .01) -> str:
    r"""
    Prints args seperated by sep with the given delay between each character and gets an input
    :param args:
    :param sep:
    :param delay:
    :return: The received input string
    """

    if not args:
        text = ''
    else:
        text = sep.join(map(str, args))

    if text:
        better_print(text, delay=delay, end='')
    res = input()
    return res



from collections.abc import Callable

from .Terminal.color import foreground
from ._check_params_decorator import check_params


RESET = foreground.RESET.value
numberColor = RESET + foreground.YELLOW.value
boolColor = RESET + foreground.LBLUE.value
stringColor = RESET + foreground.LGREEN.value
quotesColor = RESET + foreground.BOLD.value + foreground.CYAN.value

type Iterable = dict | list | tuple | set


def _f(x: object):
    if isinstance(x, str):
        quotes = '"' if "\n" not in x else '"""'
        return (f"{quotesColor}{quotes}{stringColor}{x.replace("\"", "\\\"")}"
                f"{quotesColor}{quotes}{RESET}")
    elif isinstance(x, (bool, type(None))):
        return f"{boolColor}{x}{RESET}"
    elif isinstance(x, (int, float)):
        return f"{numberColor}{x}{RESET}"
    return x


def _dictOut(_dict: dict, indentationLevel: int, indentationString: str) -> str:
    indent = indentationString * indentationLevel
    output = ""
    for k, v in _dict.items():
        if isinstance(v, (dict, list, tuple, set)):
            output += f"{indent}{_f(k)}: "
            output += _iterableOut(
                v, indentationLevel, indentationString,
                indentBefore=False
            )
        else:
            output += f"{indent}{_f(k)}: {_f(v)}\n"
    return output


def _listOut(_list: list, indentationLevel: int, indentationString: str) -> str:
    indent = indentationString * indentationLevel
    output = ""
    for item in _list:
        if isinstance(item, (dict, list, tuple, set)):
            output += _iterableOut(
                item, indentationLevel, indentationString
            )
        else:
            output += f"{indent}{_f(item)}\n"
    return output


def _tupleOut(_tuple: tuple, indentationLevel: int, indentationString: str) -> str:
    return _listOut(
        list(_tuple), indentationLevel, indentationString
    )


def _setOut(_set: set, indentationLevel: int, indentationString: str) -> str:
    return _listOut(
        list(_set), indentationLevel, indentationString
    )


def _callFunc(
        brackets: str,
        func: Callable[..., str],
        iterable: Iterable,
        indentationLevel: int,
        indentationString: str,
        indentBefore: bool) -> str:
    indent = indentationString * indentationLevel
    return (f"{indent if indentBefore else ""}{brackets[0]}\n"
            f"{func(iterable, indentationLevel + 1, indentationString)}{indent}{brackets[1]}\n")


def _iterableOut(iterable: Iterable, indentationLevel: int, indentationString: str, *,
                 indentBefore: bool = True) -> str:
    output = ""
    if isinstance(iterable, dict):
        output += _callFunc(
            "{}", _dictOut, iterable, indentationLevel,
            indentationString, indentBefore
        )
    elif isinstance(iterable, list):
        output += _callFunc(
            "[]", _listOut, iterable, indentationLevel,
            indentationString, indentBefore
        )
    elif isinstance(iterable, tuple):
        output += _callFunc(
            "()", _tupleOut, iterable, indentationLevel,
            indentationString, indentBefore
        )
    elif isinstance(iterable, set):
        output += _callFunc(
            "{}", _setOut, iterable, indentationLevel,
            indentationString, indentBefore
        )
    else:
        raise ValueError(
            f"Invalid value given for 'iterable'!\nReceived type '{type(iterable)}', expected one "
            f"of the following:\n[dict, list, tuple, set]")
    return output


@check_params
def repr_iter(
        iterable: Iterable, *,
        indentationLevel: int = 0,
        indentationString: str = "  "
) -> str:
    r"""
    Turns iterables into a formatted string for printing.
    :param iterable:
    :param indentationLevel:
    :param indentationString:
    :return: The representation of the given ``iterable``
    """

    try:
        output = _iterableOut(
            iterable,
            indentationLevel,
            indentationString
        )
        return output
    except RecursionError as e:
        return ""



import platform
import sys


if platform.system() == 'Windows':
    pass
else:
    import termios
    import tty

#from .._utils import ANSICodeBase
class ANSICodeBase:
    def __init__(self, v): ...
    def __call__(self): ...

class Cursor:
    r"""A class for manipulating the terminal cursor."""

    class __ANSICodes:
        HOME = ANSICodeBase("0;0H")
        UP = ANSICodeBase("A")
        DOWN = ANSICodeBase("B")
        FORWARD = ANSICodeBase("C")
        BACKWARD = ANSICodeBase("D")
        SAVE = ANSICodeBase("s")
        RESTORE = ANSICodeBase("u")
        REQUEST_CURSOR_POS = ANSICodeBase("6n")
        HIDE = ANSICodeBase("?25l")
        SHOW = ANSICodeBase("?25h")

    def __init__(self) -> None:
        raise RuntimeError("This class should not be instantiated.")

    @classmethod
    def home(cls) -> None:
        r"""Moves the cursor to the top left of the terminal screen."""
        cls.__ANSICodes.HOME()

    @classmethod
    def up(cls, count: int = 1) -> None:
        r"""
        Moves the cursor up ``count`` times
        :param count:
        """
        for _ in range(count):
            cls.__ANSICodes.UP()

    @classmethod
    def down(cls, count: int = 1) -> None:
        r"""
        Moves the cursor down ``count`` times
        :param count:
        """
        for _ in range(count):
            cls.__ANSICodes.DOWN()

    @classmethod
    def forward(cls, count: int = 1) -> None:
        r"""
        Moves the cursor forward ``count`` times
        :param count:
        """
        for _ in range(count):
            cls.__ANSICodes.FORWARD()

    @classmethod
    def backward(cls, count: int = 1) -> None:
        r"""
        Moves the cursor backward ``count`` times
        :param count:
        """
        for _ in range(count):
            cls.__ANSICodes.BACKWARD()

    @classmethod
    def save_pos(cls) -> None:
        r"""Saves the current cursor position."""
        cls.__ANSICodes.SAVE()

    @classmethod
    def restore_pos(cls) -> None:
        r"""Restores the saved cursor position."""
        cls.__ANSICodes.RESTORE()

    @classmethod  # !!DOESN'T WORK RN!! TODO: LOOK INTO IT AND FIX IT
    def __get_pos(cls) -> tuple[int, int]:  # !!TEMPORARILY ADDED __
        r"""Returns the current cursor position as (x, y)."""
        if platform.system() == 'Windows':
            from ctypes import Structure, c_short, c_ushort, windll, byref

            class COORD(Structure):
                _fields_ = [("X", c_short), ("Y", c_short)]

            class SMALL_RECT(Structure):
                _fields_ = [("Left", c_short), ("Top", c_short),
                            ("Right", c_short), ("Bottom", c_short)]

            class CONSOLE_SCREEN_BUFFER_INFO(Structure):
                _fields_ = [("dwSize", COORD),
                            ("dwCursorPosition", COORD),
                            ("wAttributes", c_ushort),
                            ("srWindow", SMALL_RECT),
                            ("dwMaximumWindowSize", COORD)]

            csbi = CONSOLE_SCREEN_BUFFER_INFO()
            if windll.kernel32.GetConsoleScreenBufferInfo(
                windll.kernel32.GetStdHandle(-11), byref(csbi)
            ):
                X = csbi.dwCursorPosition.X + 1
                Y = csbi.dwCursorPosition.Y + 1
                return X, Y
            else:
                raise RuntimeError("Failed to get cursor position.")
        else:
            fd = sys.stdin.fileno()
            old = termios.tcgetattr(fd)

            try:
                tty.setraw(sys.stdin.fileno())
                sys.stdout.flush()
                cls.__ANSICodes.REQUEST_CURSOR_POS()

                response = ''
                while True:
                    c = sys.stdin.read(1)
                    if c == 'R':
                        break
                    response += c

                if response.startswith('\033['):
                    response = response[2:]
                    y, x = response.split(';')
                    return int(x), int(y)
                else:
                    raise ValueError(f"Failed to get cursor position. \nResponse: {response}")

            finally:
                termios.tcsetattr(fd, termios.TCSADRAIN, old)

    @classmethod
    def set_pos(cls, x: int, y: int) -> None:
        r"""
        Sets the cursor position to (``x``, ``y``)
        :param x:
        :param y:
        """
        if x <= 0 or y <= 0:
            raise ValueError(f"Invalid cursor position. x: {x}, y: {y}\n"
                             f"The cursor starts at (1, 1).")
        ANSICodeBase(f"{y};{x}H")()

    @classmethod
    def hide(cls) -> None:
        r"""Hides the cursor."""
        cls.__ANSICodes.HIDE()

    @classmethod
    def show(cls) -> None:
        r"""Shows the cursor."""
        cls.__ANSICodes.SHOW()



from typing import Self

from .._output_stdout import output_stdout


class ANSICodeBase:
    def __init__(self, *values: int | str) -> None:
        if not values:
            raise ValueError("ANSICodeBase must be initialized with at least one value.")
        elif len(values) == 1 and isinstance(values[0], str):
            value: str = f"\033[{values[0]}"
            ansi_numbers: list[int] = []
        elif all(isinstance(value, int) for value in values):
            value: str = f"\033[{";".join(map(str, values))}m"
            ansi_numbers: list[int] = list(values)
        else:
            raise TypeError("ANSICodeBase can only be initialized with integers or a string.")

        self.value: str = value
        self.ansi_numbers: list[int] = ansi_numbers

    def __repr__(self) -> str:
        return self.value

    def __add__(self, other: str | Self) -> str | Self:
        if isinstance(other, str):
            return f"{self.value}{other}"
        elif isinstance(other, ANSICodeBase):
            if not other.ansi_numbers or not self.ansi_numbers:
                raise TypeError(
                    f"Can only add ANSICodeBase objects with ANSI number codes together."
                )
            return ANSICodeBase(*(self.ansi_numbers + other.ansi_numbers))  # problem
        else:
            raise TypeError(
                f"Cannot add value of type {type(other).__name__} to an ANSI code object."
            )

    def __radd__(self, other: str | Self) -> str | Self:
        if isinstance(other, str):
            return f"{other}{self.value}"
        elif isinstance(other, ANSICodeBase):
            if not other.ansi_numbers or not self.ansi_numbers:
                raise TypeError(
                    f"Can only add ANSICodeBase objects with ANSI number codes together."
                )
            return ANSICodeBase(*(other.ansi_numbers + self.ansi_numbers))  # problem
        else:
            raise TypeError(
                f"Cannot add value of type {type(other).__name__} to an ANSI code object."
            )

    def __call__(self) -> None:
        output_stdout(self.value)


import pyperclip

from ._check_params_decorator import check_params


@check_params
def append_clipboard(text: str) -> str:
    r"""
    Appends the given ``text`` to the clipboard
    :param text:
    :return: The given ``text``
    """
    pyperclip.copy(text)
    return text


class KWArgsHandler[_VT]:
    type __allowedKeywordsType = str | \
                                 tuple[str, type | tuple[type, ...]] | \
                                 tuple[str, type | tuple[type, ...], object]

    def __init__(
            self,
            kwargs: dict[str, _VT],
            *allowedKeywords: __allowedKeywordsType
    ) -> None:
        r"""
        Handles Keyword Arguments for a function
        :param kwargs:
        :param allowedKeywords:
        :raises ValueError: Invalid format for 'allowedKeywords' or disallowed keyword in 'kwargs'
        :raises TypeError: Invalid value type for keyword in 'kwargs'
        """
        allowedKeywords_: dict[str, tuple[type | tuple[type, ...], object]] = {}
        self.kwargs = {}
        for kw in allowedKeywords:
            if isinstance(kw, tuple) and len(kw) in (2, 3):
                keyword, argType = kw[:2]
                defValue = kw[2] if len(kw) == 3 else None
                allowedKeywords_[keyword] = (argType, defValue)
            elif isinstance(kw, str):
                allowedKeywords_[kw] = (object, None)
            else:
                raise ValueError('Invalid format for allowedKeywords. \n'
                                 'Use either ("keyword", type, defaultValue) ("keyword", type) or "keyword".')
        for keyword, value in kwargs.items():
            if keyword not in allowedKeywords_:
                raise ValueError(f"Invalid keyword '{keyword}' provided.")
            if not isinstance(value, expVT := allowedKeywords_[keyword][0]):
                raise TypeError(
                    f"Invalid type for keyword '{keyword}'. Expected {
                    expVT.__name__
                    if not isinstance(expVT, tuple)
                    else f"({" | ".join([vt.__name__ for vt in expVT])})"
                    }, got {type(value).__name__}"
                )

        for keyword, (valueType, defValue) in allowedKeywords_.items():
            value = kwargs.get(keyword, defValue)
            if keyword not in allowedKeywords_:
                raise ValueError(f"Invalid keyword '{keyword}' provided.")
            if valueType is not None and not isinstance(value, valueType):
                raise TypeError(
                    f"Invalid type for keyword '{keyword}'. Expected {
                    valueType.__name__ if not isinstance(valueType, tuple) else (vt.__name__ for vt in valueType)
                    }, got {type(value).__name__}.")
            self.kwargs[keyword] = value

    def __getattr__(self, attr) -> _VT:
        if attr in self.kwargs:
            return self.kwargs[attr]
        raise AttributeError(f"'KWArgsHandler' object has no attribute '{attr}'.")

    def __getitem__(self, item):
        if isinstance(item, int):
            raise IndexError("KWArgsHandler does not support indexing!")
        elif not isinstance(item, str):
            raise KeyError("KWArgsHandler only supports keys of type 'str'!")
        if item in self.kwargs:
            return self.kwargs[item]
        raise KeyError(f"KWArgsHandler object has no attribute '{item}'")

    def __repr__(self) -> str:
        return f"KWArgsHandler: {self.kwargs}"


from ...._utils import ANSICodeBase

BOLD = ANSICodeBase(1)
UNDERLINE = ANSICodeBase(4)
STRIKETHROUGH = ANSICodeBase(9)
SWAPCOLOR = ANSICodeBase(7)
BLACK = ANSICodeBase(30)
RED = ANSICodeBase(31)
GREEN = ANSICodeBase(32)
YELLOW = ANSICodeBase(33)
BLUE = ANSICodeBase(34)
MAGENTA = ANSICodeBase(35)
CYAN = ANSICodeBase(36)
WHITE = ANSICodeBase(37)
DEFAULT = ANSICodeBase(39)
LBLACK = ANSICodeBase(90)
LRED = ANSICodeBase(91)
LGREEN = ANSICodeBase(92)
LYELLOW = ANSICodeBase(93)
LBLUE = ANSICodeBase(94)
LMAGENTA = ANSICodeBase(95)
LCYAN = ANSICodeBase(96)
LWHITE = ANSICodeBase(97)

RESET = ANSICodeBase(0)
FGRESET = ANSICodeBase(39)


from .Terminal.color import foreground
from ._better_input import better_input
from ._check_params_decorator import check_params
from ._utils import ANSICodeBase


@check_params
def better_color_input(
        *args,
        sep: str = ' ',
        delay: float = .01,
        beforeColor: ANSICodeBase = foreground.RESET,
        inputColor: ANSICodeBase = foreground.CYAN
) -> str:
    r"""
    Prints ``args`` seperated by ``sep`` with the given ``delay`` between each character and gets an
    input in ``inputColor``. Resets the color to ``beforeColor`` afterwards
    :param args:
    :param sep:
    :param delay:
    :param beforeColor:
    :param inputColor:
    :return: The received input string
    """

    if not args:
        text: str | None = None
    else:
        text = sep.join(map(str, args))

    if text is not None:
        res = better_input(text + inputColor, delay=delay)
    else:
        res = better_input(inputColor, delay=delay)

    print(beforeColor.value, end="")

    return res


# -*- coding: utf-8 -*-

"""
Terminal
~~~~~~~~

Contains Stuff for the Terminal
"""
import platform as __platform

if __platform.system() == "Windows":
    from ._choice_interface import ChoiceInterface

from ._cursor import Cursor
from ._manager import Manager


# -*- coding: utf-8 -*-

"""
Utility for use by package-internal modules
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Contains functions, classes etc. for use by the package-internal modules.
"""

from ._ansi_code_base import ANSICodeBase


# -*- coding: utf-8 -*-

"""
Foreground
~~~~~~~~~~
ANSI codes to change the foreground (text) color
"""

from ._ansii_codes import (
    BLACK, BLUE, BOLD, CYAN, DEFAULT, FGRESET, GREEN, LBLACK, LBLUE, LCYAN, LGREEN, LMAGENTA, LRED,
    LWHITE, LYELLOW, MAGENTA, RED, RESET, STRIKETHROUGH, SWAPCOLOR, UNDERLINE, WHITE, YELLOW
)


from ...._utils import ANSICodeBase

SWAPCOLOR = ANSICodeBase(7)
BLACK = ANSICodeBase(40)
RED = ANSICodeBase(41)
GREEN = ANSICodeBase(42)
YELLOW = ANSICodeBase(43)
BLUE = ANSICodeBase(44)
MAGENTA = ANSICodeBase(45)
CYAN = ANSICodeBase(46)
WHITE = ANSICodeBase(47)
DEFAULT = ANSICodeBase(49)
LBLACK = ANSICodeBase(100)
LRED = ANSICodeBase(101)
LGREEN = ANSICodeBase(102)
LYELLOW = ANSICodeBase(103)
LBLUE = ANSICodeBase(104)
LMAGENTA = ANSICodeBase(105)
LCYAN = ANSICodeBase(106)
LWHITE = ANSICodeBase(107)

RESET = ANSICodeBase(0)
BGRESET = ANSICodeBase(49)


import random
from os import system
from string import digits

from pygetwindow import getActiveWindowTitle
from pynput.keyboard import Key, KeyCode, Listener

from .color import background, foreground
from .._utils import ANSICodeBase


type _key = Key | KeyCode
type _keys = list[_key] | tuple[_key, ...]


def clear() -> None:
    system("cls")


class ChoiceInterface:
    def __init__(
            self, *,
            textColor: ANSICodeBase = foreground.WHITE,
            highlightTextColor: ANSICodeBase = foreground.BLACK,
            highlightBackgroundColor: ANSICodeBase = background.WHITE,
            confirmKey: _keys | _key = (Key.enter, Key.right),
            cancelKey: _keys | _key = Key.esc,
            navigatePreviousKeys: _keys | _key = Key.up,
            navigateNextKeys: _keys | _key = Key.down,
            choicesSurround: str = "",
            addArrowToSelected: bool = False,
            choicesOnLine: int = 1,
            seperator: str = ""
    ) -> None:
        r"""
        Creates an instance of the interface
        :param textColor:
        :param highlightTextColor:
        :param highlightBackgroundColor:
        :param confirmKey:
        :param cancelKey:
        :param navigatePreviousKeys:
        :param navigateNextKeys:
        :param choicesSurround:
        :param addArrowToSelected:
        :param choicesOnLine:
        :param seperator:
        :raises ValueError: if values in confirmKey and cancelKey overlap
        """
        cfKeys = self.__conv_keys(confirmKey)
        ccKeys = self.__conv_keys(cancelKey)
        npKeys = self.__conv_keys(navigatePreviousKeys)
        nnKeys = self.__conv_keys(navigateNextKeys)

        seenKeys = set()
        for key in cfKeys + ccKeys + npKeys + nnKeys:
            if key in seenKeys:
                raise ValueError("Parameters specifying Keys may not overlap!")
            seenKeys.add(key)

        self.confirmKeys = cfKeys
        self.cancelKeys = ccKeys
        self.navPrevKeys = npKeys
        self.navNextKeys = nnKeys
        self.textColor = textColor
        self.hlColor = highlightTextColor + highlightBackgroundColor
        self.choicesSurround = choicesSurround
        self.addArrowToSelected = addArrowToSelected
        self.choicesOnLine = choicesOnLine
        self.seperator = seperator
        self.terminalWindowTitle = ("Choice Interface {" +
                                    "".join(random.choices(digits, k=10)) + "}")
        self.lastKeyPressed = None

    @staticmethod
    def __conv_keys(item: tuple[_key, ...] | list[_key] | _key) -> list[_key]:
        if isinstance(item, tuple):
            return list(item)
        elif isinstance(item, list):
            return item
        return [item]

    def __call__(
            self,
            choices: list[str],
            prefix: str = "",
            suffix: str = "",
            selected: int = 0,
            minimumHighlightLength: int = 0,
            terminalTitleBefore: str = "Terminal",
            returnLine: bool = False
    ) -> int | tuple[int, str]:
        r"""
        Starts the interface
        :param choices:
        :param prefix:
        :param suffix:
        :param selected:
        :param minimumHighlightLength:
        :param terminalTitleBefore:
        :param returnLine:
        :return: The index of the chosen line, or, if ``returnLine`` is ``True``, a tuple of the
        index and the line itself.
        :raises ValueError: If an argument is invalid.
        :raises Exception: If an unexpected and unaccounted-for error occurs.
        """
        system(f"TITLE {self.terminalWindowTitle}")

        if len(choices) <= 1 or (not (isinstance(choices, list) and
                                      all([isinstance(x, str) for x in choices]))):
            raise ValueError("Parameter 'lines' must be of length >= 2 and of type list[str]")
        if 0 > selected >= len(choices):
            raise ValueError(
                "Parameter 'selected' must be index of line in 'lines' and may therefore not "
                "be bigger than the biggest index of 'lines' or smaller than 0"
            )

        if minimumHighlightLength > 0:
            hlLen = minimumHighlightLength
        else:
            hlLen = max([len(line) for line in choices]) + abs(minimumHighlightLength)
            if self.addArrowToSelected:
                hlLen += 3

        while True:
            clear()
            if prefix:
                print(self.textColor + prefix + foreground.RESET)

            _out = []
            for i, line in enumerate(choices):
                if i == selected:
                    if not any([self.choicesSurround, self.addArrowToSelected]):
                        _out.append(f"{self.hlColor}{line:<{hlLen}}{foreground.RESET}")
                    elif self.addArrowToSelected:
                        _out.append(
                            f"{self.hlColor}{line:<{hlLen - 3}} > "
                            f"{foreground.RESET}"
                        )
                    else:
                        _out.append(f"{self.hlColor}{line:<{hlLen}}{foreground.RESET}")
                else:
                    _out.append(f"{self.textColor}{line:<{hlLen}}{foreground.RESET}")
                if self.choicesSurround:
                    _out.append(self.choicesSurround + _out + self.choicesSurround)

            out = []

            for i, line in enumerate(_out):
                if i % self.choicesOnLine == 0:
                    out.append("\n")
                out.append(line)
                if (i + 1) % self.choicesOnLine != 0:
                    out.append(self.seperator)

            outputstring = "".join(out)
            print(outputstring)

            if suffix:
                print(self.textColor.value + prefix + foreground.RESET)

            key: _key | None = self._waitForKey()

            if key in self.navNextKeys and selected != len(choices) - 1:
                selected += 1
            elif key in self.navPrevKeys and selected != 0:
                selected -= 1
            elif key in self.confirmKeys:
                if key == Key.enter: input()
                system(f"TITLE {terminalTitleBefore}")
                if returnLine:
                    return selected, choices[selected]
                return selected
            elif key in self.cancelKeys:
                if key == Key.enter: input()
                system(f"TITLE {terminalTitleBefore}")
                if returnLine:
                    return -1, ""
                return -1
            elif key in self.navPrevKeys + self.navNextKeys:
                pass
            else:
                raise Exception("Somehow, Somewhere, Something went wrong :/")

    def _waitForKey(self) -> _key | None:
        lst = Listener(on_press=lambda key: self._onKeyPress(key, lst))
        lst.start()
        lst.join()
        return self.__lastKeyPressed

    def _onKeyPress(self, key: _key | None, lst: Listener) -> None:
        if getActiveWindowTitle() != self.terminalWindowTitle:
            return
        self.__lastKeyPressed = key
        validKeyList: list[_key] = (
                self.confirmKeys
                + self.cancelKeys
                + self.navPrevKeys
                + self.navNextKeys
        )
        if self.__lastKeyPressed in validKeyList:
            lst.stop()


from collections.abc import Callable
from functools import wraps
from inspect import Parameter, signature
from types import UnionType
from typing import Literal, Type, TypeAliasType, Union, get_args, get_origin


def check_params[R](func: Callable[..., R]) -> Callable[..., R]:
    r"""
    Checks that all given parameters are the expected type.
    :param func:
    :return: The decorated function.
    :raises TypeError: Parameter's value is not of the expected type.
    """

    def compare_types(type1: Type, type2: Type) -> bool:
        def inner_compare_types(_type1: Type, _type2: Type) -> bool:
            _origin1, _args1 = get_origin(_type1), get_args(_type1)
            _origin2, _args2 = get_origin(_type2), get_args(_type2)
            if _origin1 is None and _origin2 is None:
                return _type1 == _type2
            if _origin1 is not None and _origin2 is not None:
                if _origin1 != _origin2: return False
                if not _args1 and not _args2: return True
                if len(_args1) != len(_args2): return False
                return all(
                    inner_compare_types(a1, a2)
                    for a1, a2 in zip(_args1, _args2)
                )
            return False

        def compare_literals(
                literals1: tuple[object, ...],
                literals2: tuple[object, ...]
        ) -> bool: return set(literals1) == set(literals2)

        if isinstance(type1, TypeAliasType):
            type1 = type1.__value__
        if isinstance(type2, TypeAliasType):
            type2 = type2.__value__

        origin1, args1 = get_origin(type1), get_args(type1)
        origin2, args2 = get_origin(type2), get_args(type2)

        if origin1 is Literal and origin2 is Literal:
            return compare_literals(args1, args2)

        return inner_compare_types(type1, type2)

    def validate_type(value: object, annot: Type) -> bool:
        if isinstance(annot, TypeAliasType):
            annot = annot.__value__

        origin = get_origin(annot)
        args = get_args(annot)

        if origin is None:
            return isinstance(value, annot)

        elif origin is tuple:
            if not isinstance(value, tuple):
                return False
            if not args: return True
            t_value: tuple = value
            if len(args) != len(t_value) and not (
                    len(args) == 2 and args[1] is Ellipsis and
                    (isinstance(args[0], UnionType) or get_origin(args[0]) is Union)
            ): return False
            elif len(args) == 2 and args[1] is Ellipsis:
                if get_origin(args[0]) is Union:
                    return all(validate_type(v, args[0]) for v in t_value)
                elif isinstance(args[0], UnionType):
                    return any(validate_type(v, args[0]) for v in t_value)
                else:
                    return all(validate_type(v, args[0]) for v in t_value)
            elif len(args) != len(t_value):
                return False
            return all(validate_type(v, args[i]) for i, v in enumerate(t_value))

        elif origin is list:
            if not isinstance(value, list):
                return False
            if not args: return True
            l_value: list = value
            return all(validate_type(v, args[0]) for v in l_value)

        elif origin is set:
            if not isinstance(value, set):
                return False
            if not args: return True
            s_value: set = value
            return all(validate_type(v, args[0]) for v in s_value)

        elif origin is dict:
            if not isinstance(value, dict):
                return False
            if not args: return True
            d_value: dict = value
            k_check: bool = all(validate_type(k, args[0]) for k in d_value.keys())
            v_check: bool = all(validate_type(v, args[1]) for v in d_value.values())
            return k_check and v_check

        elif origin is Callable:
            if not callable(value):
                return False
            if not args: return True
            if args == (Ellipsis, Ellipsis): return True
            c_value: Callable = value
            c_sig = signature(c_value)
            c_params = c_sig.parameters
            c_annots = [param.annotation for param in c_params.values()]
            c_return = c_sig.return_annotation
            expected_param_types = args[0]
            expected_return_type = args[1]
            if expected_param_types is not Ellipsis:
                if len(c_params) != len(expected_param_types):
                    return False
                if not all(
                        compare_types(v, t) for v, t in zip(
                            c_annots, expected_param_types
                        )
                ): return False
            if expected_return_type is not Ellipsis:
                if not compare_types(c_return, expected_return_type):
                    return False
            return True

        elif origin is Literal:
            return value in args

        elif origin is Union or origin is UnionType:
            return any(validate_type(value, arg) for arg in args)

        raise NotImplementedError(
            f"You stumbled upon a type that has not yet been implemented!\n"
            f"Annotation: {annot}\n"
            f"Type: {type(annot)}\n"
            f"Origin: {origin}\n"
            f"Args: {args}\n"
            f"Value: {value}"
        )

    @wraps(func)
    def wrapper(*args, **kwargs) -> R:
        params = signature(func).parameters
        for arg, param in zip(args, params.values()):
            annot = param.annotation
            if annot != Parameter.empty and not validate_type(arg, annot):
                raise TypeError(
                    f"Invalid type for parameter {param.name}: {type(arg)}. Expected: {annot}"
                )

        kwargs_param: Parameter | None = None
        for param in params.values():
            if param.kind == param.VAR_KEYWORD:
                kwargs_param = param
                break

        for k, v in kwargs.items():
            if k not in params:
                if kwargs_param is None:
                    raise TypeError(f"Got an unexpected keyword argument '{k}'")
                if kwargs_param.annotation == Parameter.empty:
                    continue
                if not validate_type(v, kwargs_param.annotation):
                    raise TypeError(
                        f"Invalid type for keyword argument: {type(v)}. Expected:"
                        f" {kwargs_param.annotation}.\nUsed key: {k}"
                    )
                continue
            annot = params[k].annotation
            if annot != Parameter.empty and not validate_type(v, annot):
                raise TypeError(
                    f"Invalid type for parameter '{k}': {type(v)}. Expected: {annot}"
                )

        return func(*args, **kwargs)

    return wrapper


import time

from ._check_params_decorator import check_params


@check_params
def better_print(*args, sep: str = ' ', delay: float = .01, end: str = '\n') -> None:
    r"""
    Prints args seperated by sep with the given delay between each character
    :param args:
    :param sep:
    :param delay:
    :param end:
    :return:
    """

    if not args:
        text = ''
    else:
        text = sep.join(map(str, args))

    for char in text:
        print(char, end='', flush=True)
        time.sleep(delay)
    print(end)


from .Terminal.color import foreground
from ._check_params_decorator import check_params
from ._utils import ANSICodeBase


@check_params
def color_input(
        *args,
        sep: str = ' ',
        beforeColor: ANSICodeBase | None = foreground.FGRESET,
        inputColor: ANSICodeBase = foreground.CYAN
) -> str:
    r"""
    Gets an input in CYAN. Resets the color to `beforeColor` afterwards
    :param args:
    :param sep:
    :param beforeColor:
    :param inputColor:
    :return:
    """
    if args:
        res = input(sep.join(map(str, args)) + inputColor)
    else:
        res = input(foreground.CYAN)
    if beforeColor is not None:
        beforeColor()
    return res


# -*- coding: utf-8 -*-

"""
Background
~~~~~~~~~~
ANSI codes to change the background color
"""

from ._ansii_codes import (
    BGRESET, BLACK, BLUE, CYAN, DEFAULT, GREEN, LBLACK, LBLUE, LCYAN, LGREEN, LMAGENTA, LRED,
    LWHITE, LYELLOW, MAGENTA, RED, RESET, SWAPCOLOR, WHITE, YELLOW
)


class kdict(dict):
    r"""Like a dict but accessible with x.key and x["key"]"""

    def __getattr__(self, name: str):
        return self[name]

    def __setattr__(self, name, value) -> None:
        self[name] = value

    def __delattr__(self, name) -> None:
        del self[name]

    def convert_inner_dicts(self, recursive: bool = False) -> None:
        for k, v in self.items():
            if isinstance(v, dict) and not isinstance(v, kdict):
                self[k] = kdict(v)
                if recursive:
                    self[k].convert_inner_dicts(True)


import sys


def output_stdout(outputValue):
    r"""
    Outputs the value to the standard output
    :param outputValue:
    :return:
    """

    sys.stdout.write(str(outputValue))
    sys.stdout.flush()


from .._utils import ANSICodeBase

class Manager:
    r"""A class for manipulating the terminal."""

    class __ANSICodes:
        CLEAR_SCREEN = ANSICodeBase("2J")
        CLEAR_CURSOR_TO_SCREEN_END = ANSICodeBase("0J")
        CLEAR_CURSOR_TO_SCREEN_START = ANSICodeBase("1J")
        CLEAR_LINE = ANSICodeBase("2K")
        CLEAR_CURSOR_TO_LINE_END = ANSICodeBase("0K")
        CLEAR_CURSOR_TO_LINE_START = ANSICodeBase("1K")

    @classmethod
    def clear_screen(cls) -> None:
        r"""Clears the terminal screen."""
        cls.__ANSICodes.CLEAR_SCREEN()

    @classmethod
    def clear_cur2s_end(cls) -> None:
        r"""Clears from the cursor to the end of the screen."""
        cls.__ANSICodes.CLEAR_CURSOR_TO_SCREEN_END()

    @classmethod
    def clear_cur2s_start(cls) -> None:
        r"""Clears from the cursor to the start of the screen."""
        cls.__ANSICodes.CLEAR_CURSOR_TO_SCREEN_START()

    @classmethod
    def clear_line(cls) -> None:
        r"""Clears the current line."""
        cls.__ANSICodes.CLEAR_LINE()

    @classmethod
    def clear_cur2l_end(cls) -> None:
        r"""Clears from the cursor to the end of the current line."""
        cls.__ANSICodes.CLEAR_CURSOR_TO_LINE_END()

    @classmethod
    def clear_cur2l_start(cls) -> None:
        r"""Clears from the cursor to the start of the current line."""
        cls.__ANSICodes.CLEAR_CURSOR_TO_LINE_START()


from typing import overload, Self


class ANSICodeBase:
    @overload
    def __init__(self, *values: int): ...
    @overload
    def __init__(self, values: str): ...
    def __repr__(self) -> str: ...
    def __add__(self, other: str | Self) -> str | Self: ...
    def __radd__(self, other: str | Self) -> str | Self: ...
    def __call__(self) -> None:

    value: str
    ansi_numbers: list[int]


# -*- coding: utf-8 -*-

"""
Maxi's Functions
~~~~~~~~~~~~~~~~

Contains all my Functions.

"""

from ._append_clipboard import append_clipboard as appendClipboard
from ._better_color_input import better_color_input as betterColorInput
from ._better_input import better_input as betterInput
from ._better_print import better_print as betterPrint
from ._check_params_decorator import check_params as checkParams
from ._color_input import color_input as colorInput
from ._key_dict import kdict as kdict
from ._kwargs_handler import KWArgsHandler as KWArgsHandler
from ._output_stdout import output_stdout as outputStdout
from ._repr_iterable import repr_iter as reprIter
